/*
** nester - NES emulator
** Copyright (C) 2000  Darren Ranalli
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful, 
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
** Library General Public License for more details.  To obtain a 
** copy of the GNU Library General Public License, write to the Free 
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
*/

#include "NES_6502.h"
#include "../NES.h"
#include <stdio.h>
#include "../../debug/debug.h"

#include "../../GBA_globals.h"

#ifndef ASSERT
#define ASSERT(EXPR)
#endif

void NES_write(uint32 address, uint8 value)
{
  globals->NES_6502_nes->MemoryWrite(address, value);
}

uint8 NES_read(uint32 address)
{
  return globals->NES_6502_nes->MemoryRead(address);
}

//4294967296 == -1 (unsigned int)
NESTER_STATIC_VAR_EXPL_Q NESTER_VAR_CONST nes6502_memread NESReadHandler[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0xFFFF, NES_read },
   { (unsigned int)-1,     (unsigned int)-1,     0 }
};

NESTER_STATIC_VAR_EXPL_Q NESTER_VAR_CONST nes6502_memwrite NESWriteHandler[] =
{
   /* $0 - $7FF is RAM */
   { 0x0800, 0xFFFF, NES_write },
   { (unsigned int)-1,     (unsigned int)-1,     0}
};

/*
NES_6502::NES_6502(NES* parent) : ParentNES(parent)
{
  if(globals->NES_6502_nes) throw "error: multiple NES_6502's";

  try {
    globals->NES_6502_nes = this;

    Init();
  } catch(...) {
    globals->NES_6502_nes = 0;
    throw;
  }
}

NES_6502::~NES_6502()
{
  globals->NES_6502_nes = 0;
}
*/

void NES_6502::FakeConstructor(NES *parent)
{
	if (globals->NES_6502_nes)
	{
		return;
	}
	globals->NES_6502_nes = this;
	ParentNES = parent;
}

// Context get/set
void NES_6502::SetContext(Context *cpu)
{
  ASSERT(0x00000000 == (cpu->pc_reg & 0xFFFF0000));
  cpu->read_handler = NESReadHandler;
  cpu->write_handler = NESWriteHandler;
  nes6502_setcontext(cpu);
}

void NES_6502::GetContext(Context *cpu)
{
  nes6502_getcontext(cpu);
  cpu->read_handler = NESReadHandler;
  cpu->write_handler = NESWriteHandler;
}

uint8 NES_6502::MemoryRead(uint32 addr)
{
  return ParentNES->MemoryRead(addr);
}

void NES_6502::MemoryWrite(uint32 addr, uint8 data)
{
  ParentNES->MemoryWrite(addr, data);
}
